(function(){

	/*
	 * This function does the interesting work,  Given an
	 * API Key, we find all Dell computers for which a
	 * BIOS serial number is known, but no ship date is
	 * known.  For each such computer, we send a request
	 * to Dell for the information.  The status of each
	 * request is written to the console.  If the API Key
	 * is rejected by Dell, this function returns false.
	 * otherwise, if the API Key did not fail, true is
	 * returned regardless of the success or failure to
	 * get the ship date of the individual computers.
	 *
	 * The ship date is stored in the a custom field
	 * in a computer record.  This must be set up in
	 * KeyConfigure in order for the field to appear.
	 */
	function process (params) {
		if (!params.apikey)
			return(false);

		app.progress.show();

		var result,
			dells = app.selection;
		if (!dells || !dells.length) {
			dells = new KSView("computer");
			dells.select("('dell'~=bman)&&(biid!=NULL)");
		}

		if (!dells.length) {
			console.info("no computers match the requirements");
			params.record && print("[info] no computers match the requirements");
			return(true);
		}

		if (params.keysecret)
			result = process_v5(dells, params);
		else
			result = process_v4(dells, params);
		if (!result)
			return(false);

		console.info("processed " + result.count + " computers (" + dells.length + " candidates)");
		params.record && print("[info] processed " + result.count + " computers (" + dells.length + " candidates)");
		if (result.ships) {
			console.info("set ship date on " + result.ships + " computers");
			params.record && print("[info] set ship date on " + result.ships + " computers");
		}
		if (result.warrants) {
			console.info("set warranty date on " + result.warrants + " computers");
			params.record && print("[info] set warranty date on " + result.warrants + " computers");
		}
		return(result.success);
	}

	function process_v5 (dells, params) {

		var xhr = new XMLHttpRequest();
		xhr.open("POST", "https://apigtwb2c.us.dell.com/auth/oauth/v2/token", false);
		xhr.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
		xhr.setRequestHeader("Accept", "application/json");
		xhr.send("grant_type=client_credentials&client_id=" + params.apikey + "&client_secret=" + params.keysecret);

		if (xhr.status != 200) {
			console.error("API key or secret is invalid (" + (xhr.statusText || xhr.status) + ")");
			params.record && print("[error] API key or secret is invalid (" + (xhr.statusText || xhr.status) + ")");
			params.record && print(xhr.responseText);
			return(false);
		}

		var success = true,
			count = 0, ships = 0, warrants = 0,
			token = xhr.response && xhr.response.access_token;

		dells.forEach(function (obj) {
			var sn = obj.biosserial;

			if (!sn || !obj.biosmodel || (obj.biosmodel.toLowerCase().indexOf('dell') < 0))
				return(false);

			app.progress.comment("requesting ship date for " + sn);
			console.info("requesting ship date for " + sn);
			params.record && print("[info] requesting ship date for " + sn);

			var xhr = new XMLHttpRequest();
			xhr.open("GET", "https://apigtwb2c.us.dell.com/PROD/sbil/eapi/v5/asset-entitlements?Method=GET&servicetags=" + sn, false);
			xhr.setRequestHeader("Authorization", "Bearer " + token);
			xhr.setRequestHeader("Accept", "application/json");
			xhr.send();

			if (xhr.status == 401) {
				console.error("bearer token is invalid (" + xhr.statusText + ")");
				params.record && print("[error] bearer token is invalid (" + xhr.statusText + ")");
				success = false;
				return(true);	/* bad api key, stop processing */
			}

			count++;
			if (xhr.status != 200) {
				console.warn("no results for " + sn + " (" + xhr.statusText + ")");
				params.record && print("[warn] no results for " + sn + " (" + xhr.statusText + ")");
			} else {
				var results = xhr.response;
				if (results
					&& results.length)
				{
					var item = results[0];
					if (params.overwrite || !obj[params.shipfield || "purchasedate"]) {
						var adate = item.shipDate;
						ships += set_date_field(obj, params.shipfield, "purchasedate", parse_date(adate));
					}

					if (params.overwrite || !obj[params.warrantyfield || "warrantydate"]) {
						var entlist = item.entitlements || [],
							dates = [];
						for (var i = 0; i < entlist.length; i++) {
							var ent = entlist[i],
								sld = ent.serviceLevelDescription,
								dat = parse_date(ent.endDate);
							if (sld && dat) {
								sld = sld.toLowerCase();
								if (sld.indexOf("onsite") >= 0 || sld.indexOf("diagnosis") >= 0)
									dates.push(dat);
							}
						}
						dates.sort(function (a, b) {
							return(parseInt(b.getTime() - a.getTime()));
						});
						warrants += set_date_field(obj, params.warrantyfield, "warrantydate", dates.length && dates[0]);
					}

					obj.commit();
				}
			}
		});

		return({ success:success, count:count, ships:ships, warrants:warrants });
	}

	function process_v4 (dells, params) {
		var success = true,
			count = 0, ships = 0, warrants = 0;

		dells.forEach(function (obj) {
			var sn = obj.biosserial;

			if (!sn || !obj.biosmodel || (obj.biosmodel.toLowerCase().indexOf('dell') < 0))
				return(false);

			app.progress.comment("requesting ship date for " + sn);
			console.info("requesting ship date for " + sn);
			params.record && print("[info] requesting ship date for " + sn);

			var xhr = new XMLHttpRequest();
			xhr.open("GET", "https://api.dell.com/support/assetinfo/v4/getassetwarranty/" + sn + "?apikey=" + params.apikey, false);
			xhr.setRequestHeader("Accept", "application/json");
			xhr.send();

			if (xhr.status == 401) {
				console.error("API key is invalid (" + xhr.statusText + ")");
				params.record && print("[error] API key is invalid (" + xhr.statusText + ")");
				success = false;
				return(true);	/* bad api key, stop processing */
			}

			count++;
			if (xhr.status != 200) {
				console.warn("no results for " + sn + " (" + xhr.statusText + ")");
				params.record && print("[warn] no results for " + sn + " (" + xhr.statusText + ")");
			} else {
				var results = xhr.response;
				results = results && results.AssetWarrantyResponse && results.AssetWarrantyResponse[0];
				if (results) {
					if (params.overwrite || !obj[params.shipfield || "purchasedate"]) {
						var adate = results.AssetHeaderData && results.AssetHeaderData.ShipDate;
						ships += set_date_field(obj, params.shipfield, "purchasedate", parse_date(adate));
					}

					if (params.overwrite || !obj[params.warrantyfield || "warrantydate"]) {
						var entlist = results.AssetEntitlementData || [],
							dates = [];
						for (var i = 0; i < entlist.length; i++) {
							var ent = entlist[i],
								sld = ent.ServiceLevelDescription,
								dat = parse_date(ent.EndDate);
							if (sld && dat) {
								sld = sld.toLowerCase();
								if (sld.indexOf("onsite") >= 0 || sld.indexOf("diagnosis") >= 0)
									dates.push(dat);
							}
						}
						dates.sort(function (a, b) {
							return(parseInt(b.getTime() - a.getTime()));
						});
						warrants += set_date_field(obj, params.warrantyfield, "warrantydate", dates.length && dates[0]);
					}

					obj.commit();
				}
			}
		});

		return({ success:success, count:count, ships:ships, warrants:warrants });
	}

	function parse_date (tdate) {
		if (!tdate)
			return tdate;
		var adate = new Date(tdate);
		return(new Date(adate.getUTCFullYear(), adate.getUTCMonth(), adate.getUTCDate(), 12, 0, 0, 0));
	}

	function set_date_field (obj, fld, def, adate, record) {
		if (!adate) {
			console.warn("no " + def + " found for " + obj.biosserial);
			record && print("[warn] no " + def + " found for " + obj.biosserial);
			return(0);
		}

		console.info("got " + def + " for " + obj.biosserial + ": " + adate);
		record && print("[info] got " + def + " for " + obj.biosserial + ": " + adate);
		if (!fld || (fld == def)) {
			obj[def] = adate;
		} else {
			adate = adate.getFullYear() + '/' + ('0' + (adate.getMonth() + 1)).substr(-2) + '/' + ('0' + adate.getDate()).substr(-2);
			obj[fld] = adate;
		}
		return(1);
	}

	/*
	 * Everything below here is UI and for requesting the API Key.
	 * First we look in the local storage for the API Key, then
	 * check if it was passed in on the command line.  Otherwise
	 * we prompt for the key.  If we get a key but it fails, we
	 * prompt again, repeating until a valid key is given or the
	 * user cancels.  If a valid key is entered, we store it in
	 * the local storage for next time.
	 */

	function prompts (params, inval) {
		if (!app.interactive)
			return;

		var usrfields = [
				{ label: "Custom Field 1", value: "usr0" },
				{ label: "Custom Field 2", value: "usr1" },
				{ label: "Custom Field 3", value: "usr2" },
				{ label: "Custom Field 4", value: "usr3" },
				{ label: "Custom Field 5", value: "usr4" },
				{ label: "Custom Field 6", value: "usr5" },
				{ label: "Custom Field 7", value: "usr6" },
				{ label: "Custom Field 8", value: "usr7" },
				{ label: "Custom Field 9", value: "usr8" },
				{ label: "Custom Field 10", value: "usr9" }
			],
			nmp = {
				version: 1,
				setup: {
					fields: [
						{
							id: "apikey",
							type: "text",
							label: "Dell API Key:",
							default: params.apikey
						},
						{
							id: "keysecret",
							type: "text",
							label: "Key Secret (for v5+):",
							default: params.keysecret
						},
						{
							id: "shipfield",
							type: "menu",
							label: "Ship Date:",
							default: params.shipfield,
							options: [ { label: "Purchase Date", value: "purchasedate" } ].concat(usrfields)
						},
						{
							id: "warrantyfield",
							type: "menu",
							label: "Warranty Date:",
							default: params.warrantyfield,
							options: [ { label: "Warranty Date", value: "warrantydate" } ].concat(usrfields)
						},
						{
							id: "overwrite",
							type: "bool",
							label: "Overwrite Existing Values",
							default: params.overwrite
						},
						{
							id: "record",
							type: "bool",
							label: "Write progress to a file",
							default: params.record
						}
					]
				},
				proceed: callback
			};
		if (inval)
			nmp.setup.fields.unshift({
				type: "note",
				label: "API key is invalid"
			});
		return(nmp);
	}

	function callback (params) {
		localStorage.setItem("shipfield", params.shipfield);
		localStorage.setItem("warrantyfield", params.warrantyfield);
		localStorage.setItem("overwrite", params.overwrite);
		localStorage.setItem("record", params.record);
		localStorage.setItem("apikey", params.apikey);
		localStorage.setItem("keysecret", params.keysecret);

		if (!process(params))
			return(prompts(params, 1));

		localStorage.setItem("apikey", params.apikey);
		localStorage.setItem("keysecret", params.keysecret);
	}

	function settings (list) {
		var params = {},
			args = app.args || {};
		for (var i = 0; i < list.length; i++) {
			var key = list[i];
			params[key] = localStorage.getItem(key) || args[key];
		}
		return(params);
	}

	var params = settings(["apikey", "keysecret", "shipfield", "warrantyfield", "overwrite", "record", "prompt"]);
	if (!app.interactive || (params.apikey && !params.prompt))
		return(callback(params));

	return(prompts(params));

})();
